import abc
import asyncio
import functools
import hashlib
import inspect
import sys
from copy import copy
from dataclasses import dataclass, field
from typing import List, Generic, Union, Callable, TypeVar, Tuple, Set, Dict, Any, Awaitable

from frozendict import frozendict
from loguru import logger
from makefun import create_function
from returns.result import safe

from pinjected.di.injected_analysis import get_instance_origin
from pinjected.di.proxiable import DelegatedVar

T, U = TypeVar("T"), TypeVar("U")

A = TypeVar("A")
B = TypeVar("B")

INJECTED_CONTEXT = frozendict()
"""
This is a global context for injected.
This context is copied on the creation time of an Injected.
This is useful for passing the context to the Injected instance.
However, be careful that you pass correct context.
For example, whe using with DelegatedVar, do keep the context in DelegatedVar and then pass the context to InjectedInstance at the end.
Also, if beware that the context will not be propagated for operations.
examples:
with InjectedContext({"a":1}):
    a = Injected.by_name("a")
    b = a.map(lambda x:x+1) # here b will have the context {"a":1}, so it will return 2
b = a.map(lambda x:x+1) # here the map function will not have the context {"a":1}, so it will return undefined.
to mitigate this, we need modify all the operators to propagate the parent context, which is not easy.
Okey, so let's implement this and be aware that the map function will erase it.
So the use needs to be careful to add context at the very end, for running purposes.
"""


@dataclass
class FrameInfo:
    filename: str
    line_number: int
    function_name: str
    sources: List[str]
    line_idx_in_sources: int


def get_frame_info(stack_idx) -> Union[None, FrameInfo]:
    try:
        return FrameInfo(*inspect.getframeinfo(inspect.stack(0)[stack_idx][0]))
    except Exception as e:
        return None


def partialclass(name, cls, *args, **kwds):
    """
    copied from https://stackoverflow.com/questions/38911146/python-equivalent-of-functools-partial-for-a-class-constructor
    :param name:
    :param cls:
    :param args:
    :param kwds:
    :return:
    """
    new_cls = type(name, (cls,), {
        '__init__': functools.partialmethod(cls.__init__, *args, **kwds)
    })

    # The following is copied nearly ad verbatim from `namedtuple's` source.
    """
    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython).
    """
    try:
        new_cls.__module__ = sys._getframe(1).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return new_cls


@dataclass
class ParamInfo:
    params_to_fill: Dict  # = tgt_sig.parameters
    params_state: Dict  # = dict()
    vargs: List = field(default_factory=list)  # = []
    kwargs: Dict = field(default_factory=dict)  # = dict()
    # the problem is we may or may not have vargs and kwargs...


class Injected(Generic[T], metaclass=abc.ABCMeta):
    """
    The ``Injected`` class represents a sophisticated dependency injection mechanism in Python.
    It encapsulates an object that requires certain dependencies to be resolved. The class maintains
    a set of dependencies necessary for the object's creation and utilizes a provider function that
    generates the desired variable with its dependencies satisfied.

    Basic Usage:
    ------------

    .. code-block:: python

        from pinjected.di.util import Injected
        from pinjected import Design

        def provide_ab(a:int, b:int) -> int:
            return a + b

        # The bind method creates an Injected object from a provider function, using the function's
        # arguments as dependencies.
        injected: Injected[int] = Injected.bind(provide_ab)

        design = Design().bind_instance(a=1, b=2)
        assert design.to_graph()[injected] == 3

    Advanced Features:
    ------------------

    **Composition:**

    ``Injected`` instances can be manipulated and combined in several ways to build complex dependency structures.

    1. **map**: Transform the result of an ``Injected`` instance.

        .. code-block:: python

            from pinjected.di.util import Injected, instances
            from pinjected import Design

            design: Design = instances(a=1)  # Shortcut for binding instances
            a: Injected[int] = Injected.by_name('a')
            b: Injected[int] = a.map(lambda x: x + 1)  # b is now a + 1

            g = design.to_graph()
            assert g[a] + 1 == g[b]

    2. **zip/mzip**: Combine multiple ``Injected`` instances.

        .. code-block:: python

            from pinjected.di.util import Injected
            from pinjected import Design

            design = Design().bind_instance(a=1, b=2)
            g = design.to_graph()

            a = Injected.by_name('a')
            b = Injected.by_name('b')
            c = a.map(lambda x: x + 2)
            abc = Injected.mzip(a, b, c)  # Combine more than two Injected instances
            ab_zip = Injected.zip(a, b)   # Standard zip for two Injected instances

            assert g[abc] == (1, 2, 3)
            assert g[ab_zip] == (1, 2)

    3. **dict/list**: Create a dictionary or list from multiple ``Injected`` instances.

        .. code-block:: python

            from pinjected.di.util import Injected, instances

            design = instances(a=1, b=2)
            a = Injected.by_name('a')
            b = Injected.by_name('b')
            c = a.map(lambda x: x + 2)

            injected_dict: Injected[dict] = Injected.dict(a=a, b=b, c=c)  # Creates {'a':1, 'b':2, 'c':3}
            injected_list: Injected[list] = Injected.list(a, b, c)  # Creates [1, 2, 3]

    These composition tools enhance the flexibility of dependency injection, enabling more complex
    and varied structures to suit diverse programming needs. By allowing transformations and combinations
    of ``Injected`` instances, they provide powerful ways to manage and utilize dependencies within
    your applications.
    """

    @staticmethod
    def partial(original_function: Callable, **injection_targets: "Injected") -> "Injected[Callable]":
        """
        Partially injects dependencies into the parameters of the target function.

        This utility allows for specific arguments of a callable to be automatically populated with injected dependencies,
        leaving the remainder to be supplied at the point of invocation. It facilitates a cleaner separation of concerns
        by decoupling the resolution of dependencies from the function's implementation.

        Parameters:
        original_function (Callable): The original function into which dependencies are to be injected. This function maintains its original signature, with certain parameters now automatically resolved via injection.
        injection_targets (dict): A mapping of parameter names to their injected dependencies. Each entry corresponds to a specific argument of `original_function` and the dependency to be injected into it. These parameters will be automatically populated when the returned function is invoked.

        Returns:
        Injected[Callable]: A wrapped version of `original_function` that, when called, will receive the injected dependencies specified in `injection_targets`. The callable requires the remaining parameters that were not part of `injection_targets`, maintaining the order and names as in the original function's signature.

        Note:
        The implementation focuses on flexibility and separation of concerns at the cost of some complexity in the
        function's wrapping mechanics. This design choice prioritizes decoupling over simplicity to ensure that functions
        can be composed and dependencies managed outside of the business logic.

        :Example:

        .. code-block:: python

            def f(a, b, c):
                return a + b + c

            # Creating a partially injected function
            injected_func = Injected.partial(f, a=1, b=2)  # 'a' and 'b' are injected, 'c' is left to be provided later.

            # Using the partially injected function
            result:Injected[int] = injected_func(c=3)  # Now we provide 'c', and the function executes with all parameters.
            g[result] == 6  # The result is 6. g[] is a syntax sugar for g.provide(result)

            # Alternatively, using different syntax or within different contexts, the following are equivalent:
            assert g[Injected.pure(6)] == g[result]  # Comparing with a pure value
            assert g[Injected.partial(f, a=1, b=2)](c=3) == 6  # Using the function within a context 'g'
            assert g[Injected.partial(f, a=1, b=2)(c=3)] == 6  # Immediate invocation within a context 'g'

        These examples illustrate how `partial` is used to pre-fill certain parameters of a function, allowing the
        remaining ones to be specified at a later point in the code or within different execution contexts.
        """
        # WARNING DO NOT EVER USE LOGGER HERE. IT WILL CAUSE PICKLING ERROR on ray's nested remote call!
        original_sig = inspect.signature(original_function)

        # USING a logger in here make things very difficult to debug. because makefun doesnt seem to keep __closure__
        # hmm we need to check if the args are positional only or not.
        # in that case we need to inject with *args.
        def _get_new_signature(funcname, missing_params):
            missing_non_defaults = [p for p in missing_params if
                                    p.default is inspect.Parameter.empty and p.kind != inspect.Parameter.VAR_KEYWORD and p.kind != inspect.Parameter.VAR_POSITIONAL]
            # hmm, we need to check if the original_funcion is a method or not.
            # if method, ignore the first param.

            vkwarg = [p for p in missing_params if p.kind == inspect.Parameter.VAR_KEYWORD]
            if not vkwarg:
                vkwarg = [inspect.Parameter('__kwargs', inspect.Parameter.VAR_KEYWORD)]
            varg = [p for p in missing_params if p.kind == inspect.Parameter.VAR_POSITIONAL]
            if not varg:
                varg = [inspect.Parameter('__args', inspect.Parameter.VAR_POSITIONAL)]
            # we also need to pass varargs if there are default args..
            new_func_sig = f"injected_{funcname}({','.join([str(p).split(':')[0] for p in (missing_non_defaults + varg + vkwarg)])})"
            return new_func_sig

        def makefun_impl(injected_kwargs):
            # this gets called every time you call this function through PartialInjectedFunction interface
            # this is because the injected_kwargs gets changed.

            missing_keys = [k for k in original_sig.parameters.keys() if k not in injected_kwargs]
            missing_params = [original_sig.parameters[k] for k in missing_keys]
            missing_positional_args = [p for p in missing_params if p.kind == inspect.Parameter.POSITIONAL_ONLY]
            missing_non_positional_args = [p for p in missing_params if
                                           p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD]
            new_func_sig = _get_new_signature(original_function.__name__, missing_params)
            defaults = {p.name: p.default for p in missing_params if p.default is not inspect.Parameter.empty}

            def func_gets_called_after_injection_impl(*_args, **_kwargs):
                assert len(_args) >= len(
                    missing_positional_args), f"not enough args for positional only args:{missing_positional_args}"
                num_missing_positional_args = len(missing_positional_args)
                inferred_pos_arg_names = [p.name for p in missing_positional_args]
                inferred_non_pos_arg_names = [p.name for p in missing_non_positional_args]
                inferred_positional_args, inferred_non_positional_args = _args[:num_missing_positional_args], _args[
                                                                                                              num_missing_positional_args:]
                # logger.info(f"inferred positional args:{inferred_positional_args}")
                # logger.info(f"inferred non positional args:{inferred_non_positional_args}")
                # logger.info(f"inferred positional arg names:{inferred_pos_arg_names}")
                # logger.info(f"inferred non positional arg names:{inferred_non_pos_arg_names}")
                total_kwargs = copy(defaults)
                filled_kwargs = {**injected_kwargs,
                                 **dict(zip(inferred_pos_arg_names, inferred_positional_args)),
                                 **dict(zip(inferred_non_pos_arg_names, inferred_non_positional_args)),
                                 **_kwargs}
                # logger.info(f"filled kwargs:{filled_kwargs}")
                total_kwargs.update(filled_kwargs)
                # logger.info(f"total kwargs:{total_kwargs}")
                args = [total_kwargs[k] for k, p in original_sig.parameters.items() if
                        p.kind != inspect.Parameter.VAR_KEYWORD and p.kind != inspect.Parameter.VAR_POSITIONAL and p.kind != inspect.Parameter.KEYWORD_ONLY]
                # we need to put the ramaining args into kwargs if kwargs is present in the signature
                vks = [p for k, p in original_sig.parameters.items() if p.kind == inspect.Parameter.VAR_KEYWORD]
                if vks:
                    kwargs = {k: v for k, v in total_kwargs.items() if k not in original_sig.parameters.keys()}
                else:
                    kwargs = {}
                # we also need to handle kwonly args
                kws = [p for k, p in original_sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY]
                if kws:
                    for k in kws:
                        kwargs[k.name] = total_kwargs[k.name]
                vas = [p for k, p in original_sig.parameters.items() if p.kind == inspect.Parameter.VAR_POSITIONAL]
                if vas:
                    # we need to put the ramaining args into kwargs if kwargs is present in the signature
                    vargs = _args[num_missing_positional_args + len(missing_non_positional_args):]
                else:
                    vargs = []
                # hmm we need to pass kwargs too..
                # logger.info(f"args:{args}")
                # args contains values from kwargs...
                # logger.info(f"vargs:{vargs}")
                # logger.info(f"kwargs:{kwargs}")
                bind_result = original_sig.bind(*args, *vargs, **kwargs)
                bind_result.apply_defaults()
                # logger.info(f"bound args:{bind_result.args}")
                # logger.info(f"bound kwargs:{bind_result.kwargs}")
                # Ah, since the target_function is async, we can't catch...
                return original_function(*bind_result.args, **bind_result.kwargs)

            # logger.info(f"injected.partial -> {new_func_sig} ")
            new_func = create_function(
                new_func_sig,
                func_gets_called_after_injection_impl,
                doc=original_function.__doc__,
            )
            new_func.__is_async__ = inspect.iscoroutinefunction(original_function)
            __doc__ = original_function.__doc__
            __skeleton__ = f"""def {new_func_sig}:
    \"\"\"
    {__doc__}
    \"\"\"
"""
            new_func.__skeleton__ = __skeleton__

            return new_func

        makefun_impl.__name__ = original_function.__name__
        makefun_impl.__module__ = original_function.__module__
        makefun_impl.__original__ = original_function

        if isinstance(original_function, type):
            makefun_impl.__original_code__ = "not available"
            makefun_impl.__original_file__ = "not available"
        elif type(original_function).__name__ == 'staticmethod':
            makefun_impl.__original_code__ = inspect.getsource(original_function.__func__)
            makefun_impl.__original_file__ = inspect.getfile(original_function.__func__)
        else:
            makefun_impl.__original_code__ = safe(inspect.getsource)(original_function).value_or("not available")
            makefun_impl.__original_file__ = safe(inspect.getfile)(original_function).value_or("not available")

        makefun_impl.__doc__ = original_function.__doc__
        injected_kwargs = Injected.dict(**injection_targets)
        # hmm?
        # Ah, so upon calling instance.method(), we need to manually check if __self__ is present?
        injected_factory = PartialInjectedFunction(
            Injected.bind(makefun_impl, injected_kwargs=injected_kwargs)
        )
        # the inner will be called upon calling the injection result.
        # This involves many internal Injecte instances. can I make it simler?
        # it takes *by_name, mzip, and map.
        injected_factory.__runnable_metadata__ = {
            "kind": "callable"
        }

        return injected_factory

    @staticmethod
    def inject_except(target_function, *whitelist: str) -> "Injected[Callable]":
        """
        :param target_function:
        :param whitelist: name of arguments which should not be injected by DI.
        :return: Injected[Callable[(whitelisted args)=>Any]]
        """
        argspec = inspect.getfullargspec(target_function)
        args_to_be_injected = [a for a in argspec.args if a not in whitelist and a != "self"]
        return Injected.partial(target_function, **{item: Injected.by_name(item) for item in args_to_be_injected})

    @staticmethod
    def bind(_target_function_, **kwargs_mapping: Union[str, type, Callable, "Injected"]) -> "InjectedFunction":
        if isinstance(_target_function_, Injected):
            _target_function_ = _target_function_.get_provider()
        return InjectedFunction(
            target_function=_target_function_,
            kwargs_mapping=kwargs_mapping
        )

    @staticmethod
    def direct(_target_function, **kwargs) -> "Injected":
        """
        uses Injected.pure by default unless the parameter is an instance of Injected.
        :param _target_function:
        :param kwargs:
        :return:
        """
        en_injected = {k: Injected.pure(v) for k, v in kwargs.items() if not isinstance(v, (Injected, DelegatedVar))}
        already_injected = {k: v for k, v in kwargs.items() if isinstance(v, (Injected, DelegatedVar))}
        return Injected.bind(_target_function, **en_injected, **already_injected)

    def _faster_get_fname(self):

        try:
            frame = sys._getframe().f_back.f_back.f_back.f_back
            mod = frame.f_globals["__name__"]
            name = frame.f_lineno
            return f"{mod.replace('.', '_')}_L_{name}".replace("<", "__").replace(">", "__")
        except Exception as e:
            # from loguru import logger
            # logger.warning(f"failed to get name of the injected location.")
            return f"__unknown_module__maybe_due_to_pickling__"

    def __init__(self):
        self.fname = self._faster_get_fname()

    def apply_partial(self, *args: "Injected", **kwargs: "Injected"):
        assert all(isinstance(a, Injected) for a in args), f"{args} is not all Injected"
        assert all(isinstance(a, Injected) for a in kwargs.values()), f"{kwargs} is not all Injected"
        args = Injected.mzip(*args)
        kwargs = Injected.dict(**kwargs)
        f = self
        res = Injected.mzip(f, args, kwargs).map(
            lambda f_args_kwargs: functools.partial(f_args_kwargs[0], *f_args_kwargs[1], **f_args_kwargs[2]))
        return PartialInjectedFunction(res)

    @abc.abstractmethod
    def dependencies(self) -> Set[str]:
        pass

    def dynamic_dependencies(self) -> Set[str]:
        """
        :return: a set of dependencies which are not statically known. mainly used for analysis.

        use this to express an injected that conditionally depends on something, such as caches.
        """
        return self.dependencies()

    def get_signature(self):
        sig = f"""{self.fname}({",".join(self.dependencies())})"""
        # logger.warning(sig)
        return sig

    @abc.abstractmethod
    def get_provider(self):
        """
        :return: a provider
        """

    def map(self, f: Callable[[T], U]) -> 'Injected[U]':
        return MappedInjected(self, f)

    def from_impl(impl: Callable, dependencies: Set[str]):
        return GeneratedInjected(impl, dependencies)

    @staticmethod
    def pure(value):
        return InjectedPure(value)

    @staticmethod
    def by_name(name: str):
        return InjectedByName(name, )

    def zip(self, other: "Injected[U]") -> "Injected[Tuple[T,U]]":
        assert isinstance(self, Injected)
        assert isinstance(other, Injected)
        return ZippedInjected(self, other)

    @staticmethod
    def mzip(*srcs: "Injected"):
        srcs = [Injected.ensure_injected(s) for s in srcs]
        return MZippedInjected(*srcs)

    @staticmethod
    def tuple(*srcs: "Injected"):
        return Injected.mzip(*srcs).map(lambda t: tuple(t))

    @staticmethod
    def list(*srcs: "Injected"):
        return Injected.mzip(*srcs).map(list)

    @staticmethod
    def async_gather(*srcs: "Injected[Awaitable]"):
        def gather(coros):
            async def wait():
                return await asyncio.gather(*coros)

            return asyncio.run(wait())

        return Injected.mzip(*srcs).map(gather)

    @staticmethod
    def map_elements(f: "Injected[Callable]", elements: "Injected[Iterable]") -> "Injected[Iterable]":
        """
        # for (
        #     f <- f,
        #     elements <- elements
        #     ) yield map(f,elements)

        :param f:
        :param elements:
        :return:

        """
        return Injected.mzip(
            f, elements
        ).map(lambda x: map(x[0], x[1]))

    # this is ap of applicative functor.
    def apply_injected_function(self, other: "Injected[Callable[[T],U]]") -> "Injected[U]":
        return self.zip(other).map(
            lambda t: t[1](t[0])
        )

    def and_then_injected(self, other: "Injected[Callable[[T],U]]") -> "Injected[Callable[[Any],U]]":
        return self.zip(other).map(
            lambda t: lambda *args, **kwargs: t[1](t[0](*args, **kwargs))
        )

    @staticmethod
    def dict(**kwargs: "Injected") -> "Injected[Dict]":
        keys = list(kwargs.keys())
        return Injected.mzip(*[kwargs[k] for k in keys]).map(lambda t: {k: v for k, v in zip(keys, t)})

    @property
    def proxy(self) -> DelegatedVar:
        """use this to modify injected variables freely without map.
        call eval() at the end to finish modification
        # it seems this thing is preventing from pickling?
        """
        from pinjected.di.app_injected import injected_proxy
        return injected_proxy(self)

    @staticmethod
    def ensure_injected(data: Union["Injected", DelegatedVar]):
        match data:
            case DelegatedVar():
                return Injected.ensure_injected(data.eval())
            case Injected():
                return data
            case func if callable(func):
                return Injected.bind(func)
            case _:
                raise RuntimeError(f"not an injected object: {data},type(data)={type(data)}")

    def __add__(self, other: "Injected"):
        match other:
            case int() | float() | str():
                other = Injected.pure(other)
        other = Injected.ensure_injected(other)
        return self.zip(other).map(lambda t: t[0] + t[1])

    def __getitem__(self, item):
        return self.map(lambda x: x[item])

    def desync(self):
        async def impl(awaitable):
            return await awaitable

        return self.map(lambda coroutine: asyncio.run(impl(coroutine)))

    def __len__(self):
        return self.map(len)

    # Implementing these might end up with pickling issues. due to recursive getattr..?
    # def __call__(self, *args, **kwargs):
    #     return self.proxy(*args, **kwargs)
    #
    # def __getitem__(self, item):
    #     return self.proxy[item]

    @staticmethod
    def conditional(condition: "Injected[bool]", true_case: "Injected", false_case: "Injected"):
        return ConditionalInjected(condition, true_case, false_case)

    @staticmethod
    def procedure(*targets: "Injected"):
        """
        Runs the targets in order, and returns the last one.
        This is useful for running injecteds which performs side effects.
        :param targets:
        :return:
        """
        return Injected.list(*targets).map(lambda items: items[-1])

    @staticmethod
    def conditional_preparation(
            condition: "Injected[bool]",
            preparation: "Injected",
            utilization: "Injected"
    ):
        return Injected.conditional(
            condition,
            utilization,
            Injected.procedure(
                preparation,
                utilization
            )
        )


@dataclass
class ConditionalInjected(Injected):
    condition: Injected[bool]
    true: Injected
    false: Injected

    def dependencies(self) -> Set[str]:
        return self.condition.dependencies() | {"session"}

    def get_provider(self):
        def task(condition, session: "IObjectGraph"):
            if condition:
                return session[self.true]
            else:
                return session[self.false]

        return Injected.bind(task, condition=self.condition).get_provider()

    def dynamic_dependencies(self) -> Set[str]:
        return self.condition.dynamic_dependencies() | \
            self.true.dynamic_dependencies() | \
            self.false.dynamic_dependencies() | {"session"}


@dataclass
class InjectedCache(Injected[T]):
    """
    A specialized Injected class that handles the caching of program execution results.

    The InjectedCache class is designed to manage caching within a dependency injection framework. It ensures that if a program's execution result, based on certain dependencies, is already known (cached), it does not need to be recalculated. This mechanism is particularly useful for operations that are computationally expensive or require resources that are costly to access repeatedly.

    Attributes:
    -----------
    cache : Injected[Dict]
        An Injected dictionary that represents the cache storage.

    program : Injected[T]
        The program whose result needs to be cached. This is typically a complex operation or query.

    program_dependencies : List[Injected]
        A list of dependencies required by the program. These dependencies are monitored for changes that might invalidate the cache.

    Methods:
    --------
    __post_init__(self):
        Ensures the 'program' attribute is of type 'Injected' and sets up the internal caching mechanism.

    get_provider(self):
        Retrieves the provider for the Injected instance, which is the caching mechanism itself in this case.

    dependencies(self) -> Set[str]:
        Returns the set of dependencies' names required by the caching mechanism.

    dynamic_dependencies(self) -> Set[str]:
        Computes and returns a set of dynamic dependencies by combining the cache's own dependencies and the program's dynamic dependencies.

    __hash__(self):
        Returns the hash of the 'impl' attribute, which represents the unique configuration of the cache.

    Usage:
    ------
    The InjectedCache is not typically created directly by users. Instead, it's often part of a larger dependency injection framework where caching is required. When a certain condition or set of conditions is met, the InjectedCache object checks its internal storage (the 'cache' attribute) to determine whether the current operation's result already exists. If it does, the cached result is returned, saving time and resources. Otherwise, the operation is performed, and the result is stored in the cache for future use.

    This class significantly optimizes performance, especially in scenarios where operations are repeated with the same parameters or contexts, and computational resources are scarce or expensive.
    """
    cache: Injected[Dict]
    program: Injected[T]
    program_dependencies: List[Injected]

    def __post_init__(self):
        self.program = Injected.ensure_injected(self.program)

        def impl(session, cache: Dict, *deps):
            logger.info(f"Checking for cache with deps:{deps}")
            sha256_key = hashlib.sha256(str(deps).encode()).hexdigest()
            hash_key = sha256_key
            if hash_key not in cache:
                logger.info(f"Cache miss for {deps}")
                data = session[self.program]
                cache[hash_key] = data
            else:
                logger.info(f"Cache hit for {deps},loading ...")
            res = cache[hash_key]
            logger.info(f"Cache hit for {deps}, loaded")
            return res

        self.impl = Injected.list(
            Injected.by_name("session"),
            self.cache,
            *self.program_dependencies
        ).map(
            lambda t: impl(*t)
        )
        assert isinstance(self.impl, Injected)
        assert isinstance(self.program, Injected)

    def get_provider(self):
        return self.impl.get_provider()

    def dependencies(self) -> Set[str]:
        return self.impl.dependencies()

    def dynamic_dependencies(self) -> Set[str]:
        return self.impl.dynamic_dependencies() | \
            self.program.dynamic_dependencies()

    def __hash__(self):
        return hash(self.impl)


class IAsyncDict(abc.ABC):
    @abc.abstractmethod
    async def get(self, key):
        pass

    @abc.abstractmethod
    async def set(self, key, value):
        pass

    @abc.abstractmethod
    async def delete(self, key):
        pass

    @abc.abstractmethod
    async def contains(self, key):
        pass


async def auto_await(tgt):
    if inspect.isawaitable(tgt):
        return await tgt
    else:
        return tgt


@dataclass
class AsyncInjectedCache(Injected[T]):
    """
    Represents a specialized caching mechanism within an asynchronous dependency injection system.

    This class manages the caching of results from asynchronous operations or programs. By monitoring dependencies defined in the system's design, it optimizes resource usage and performance, ensuring that results for repeated operations are reused when applicable.

    Attributes:
    -----------
    cache : Injected[IAsyncDict]
        An asynchronous dictionary acting as the cache storage.

    program : Injected[Awaitable[T]]
        The main program or operation whose results are to be cached. It is wrapped in an `Injected` to ensure compatibility with the dependency injection system.

    program_dependencies : list[Injected]
        Specific dependencies that the program relies on, which may influence the caching mechanism.

    Methods:
    --------
    __post_init__(self):
        Initializes the internal structures and ensures the 'program' is of type 'Injected'. Sets up the asynchronous caching strategy.

    get_provider(self):
        Retrieves the provider function responsible for fetching or computing the necessary data, handling caching logic in the background.

    dependencies(self) -> Set[str]:
        Identifies and returns a set of static dependencies required by the caching system.

    dynamic_dependencies(self) -> Set[str]:
        Determines and returns a set of dynamic dependencies that can change over runtime, affecting the caching mechanism.

    __hash__(self):
        Provides a unique hash representing the current state of the cache configuration, aiding in cache invalidation and recognition.

    Usage:
    ------
    The `AsyncInjectedCache` class is particularly useful in scenarios involving repetitive asynchronous operations where results can be cached to improve performance. It intercepts calls to the encapsulated 'program', checks the 'cache' for existing results, and either returns the cached data or proceeds with the operation, caching the new results. This process is seamless to the user, ensuring efficient use of resources and faster data retrieval, thanks to asynchronous processing.

    Note:
    -----
    This class requires careful handling of dependencies, especially when they are awaitable. The caching mechanism relies on the consistency and predictability of these dependencies to function correctly.
    """
    cache: Injected[IAsyncDict]
    program: Injected[Awaitable[T]]
    program_dependencies: list[Injected]

    def __post_init__(self):
        self.program = Injected.ensure_injected(self.program)
        assert isinstance(self.program, Injected)
        assert isinstance(self.program_dependencies, list), f"program_dependencies:{self.program_dependencies}"

        from pinjected.di.decorators import cached_coroutine
        @cached_coroutine
        async def impl(session, cache: IAsyncDict, *deps):
            from loguru import logger
            # deps can be awaitable, so beware...
            assert isinstance(cache, IAsyncDict)
            deps = await asyncio.gather(*[auto_await(t) for t in deps])
            logger.info(f"Checking cache for {self.program} with deps:{deps}")
            sha256_key = hashlib.sha256(str(deps).encode()).hexdigest()
            hash_key = sha256_key
            if not await cache.contains(hash_key):
                logger.info(f"Cache miss for {deps} in {cache}")
                data = await session[self.program]
                logger.info(f"Cache miss for {deps}, tried {cache}, writing...")
                await cache.set(hash_key, data)
                logger.info(f"Writen to cache for {deps} to {cache}")
            else:
                logger.info(f"Cache hit for {deps},loading from {cache}")
            res = await cache.get(hash_key)
            logger.info(f"Cache hit for {deps}, loaded from {cache}")
            return res

        deps = Injected.list(
            Injected.by_name("session"),
            self.cache,
            *self.program_dependencies,
        )
        self.impl = deps.map(lambda t: impl(*t))
        assert isinstance(self.impl, Injected)
        assert isinstance(self.program, Injected)

    def get_provider(self):
        return self.impl.get_provider()

    def dependencies(self) -> Set[str]:
        return self.impl.dependencies()

    def dynamic_dependencies(self) -> Set[str]:
        return self.impl.dynamic_dependencies() | \
            self.program.dynamic_dependencies()

    def __hash__(self):
        return hash(self.impl)


class GeneratedInjected(Injected):
    """creates Injected from dependencies and funct(**kwargs) signature"""

    def __init__(self, impl: Callable, dependencies: Set[str]):
        super().__init__()
        self.impl = impl
        self._dependencies = dependencies

    def dependencies(self) -> Set[str]:
        return self._dependencies

    def get_provider(self):
        return create_function(self.get_signature(), func_impl=self.impl)


class MappedInjected(Injected):
    __match_args__ = ("src", "f")

    def __init__(self, src: Injected[T], f: Callable[[T], U]):
        super(MappedInjected, self).__init__()
        self.src = src
        self.f: Callable[[T], U] = f

    def dependencies(self) -> Set[str]:
        return self.src.dependencies()

    def get_provider(self):
        def impl(**kwargs):
            assert self.dependencies() == set(
                kwargs.keys())  # this is fine but get_provider returns wrong signatured func
            tmp = self.src.get_provider()(**kwargs)
            return self.f(tmp)

        return create_function(self.get_signature(), func_impl=impl)


def extract_dependency_including_self(f: Union[type, Callable]):
    if isinstance(f, type):  # it's a constructor and we must use __init__ for backward compatibility.
        argspec = inspect.getfullargspec(f.__init__)
        return set(argspec.args)
    elif isinstance(f, Callable):
        argspec = inspect.getfullargspec(f)
        return set(argspec.args)
    else:
        raise RuntimeError(f"input:{f} is not either type or Callable")


def extract_dependency(dep: Union[str, type, Callable, Injected, DelegatedVar[Injected]]) -> Set[str]:
    if isinstance(dep, str):
        return {dep}
    elif isinstance(dep, type):  # it's a constructor and we must use __init__ for backward compatibility.
        argspec = inspect.getfullargspec(dep.__init__)
        return set(argspec.args) - {'self'}
    elif isinstance(dep, DelegatedVar):
        return extract_dependency(dep.eval())
    elif isinstance(dep, Injected):
        return dep.dependencies()
    elif isinstance(dep, Callable):
        try:
            argspec = inspect.getfullargspec(dep)
        except Exception as e:
            from loguru import logger
            logger.error(f"failed to get argspec of {dep}. of type {type(dep)}")
            raise e

        return set(argspec.args) - {'self'}
    else:
        raise RuntimeError(f"dep must be either str/type/Callable/Injected. got {type(dep)}")


def solve_injection(dep: Union[str, type, Callable, Injected], kwargs: dict):
    if isinstance(dep, str):
        return kwargs[dep]
    elif isinstance(dep, DelegatedVar):
        return solve_injection(dep.eval(), kwargs)
    elif isinstance(dep, Injected):
        return solve_injection(dep.get_provider(), kwargs)
    elif isinstance(dep, (type, Callable)):
        return dep(**{k: kwargs[k] for k in extract_dependency(dep)})
    else:
        raise RuntimeError(f"dep must be one of str/type/Callable/Injected. got {type(dep)}")


def combine_image_store(a, b):
    # do anything
    return a + b


def assert_kwargs_type(v):
    if isinstance(v, str):
        return
    if isinstance(v, type):
        return
    if isinstance(v, Callable):
        return
    if isinstance(v, Injected):
        return
    else:
        raise TypeError(f"{type(v)} is not any of [str,type,Callable,Injected],but {v}")


class InjectedFunction(Injected[T]):
    # since the behavior differs in classes extending Generic[T]
    __match_args__ = ("target_function", "kwargs_mapping")

    def __init__(self,
                 target_function: Callable,
                 kwargs_mapping: Dict[str, Union[str, type, Callable, Injected, DelegatedVar]]
                 ):
        # I think we need to know where this class is instantiated outside of pinjected_package
        self.origin_frame = get_instance_origin("pinjected")
        super().__init__()
        assert not isinstance(target_function, (Injected, DelegatedVar))
        assert callable(target_function)
        self.target_function = target_function
        self.kwargs_mapping = copy(kwargs_mapping)
        for k, v in self.kwargs_mapping.items():
            assert_kwargs_type(v)
            if isinstance(v, DelegatedVar):
                self.kwargs_mapping[k] = v.eval()
        # logger.info(f"InjectedFunction:{self.target_function} kwargs_mapping:{self.kwargs_mapping}")
        org_deps = extract_dependency(self.target_function)
        # logger.info(f"tgt:{target_function} original dependency:{org_deps}")
        missings = {d for d in org_deps if d not in self.kwargs_mapping}
        # logger.info(f"now missing {missings}")
        # logger.warning(f"created InjectedFunction:{inspect.signature(target_function)}")
        # assert "self" not in inspect.signature(target_function).parameters
        self.missings = missings

    def override_mapping(self, **kwargs: Union[str, type, Callable, Injected]):
        return InjectedFunction(self.target_function, {**self.kwargs_mapping, **kwargs})

    def get_provider(self):
        signature = self.get_signature()

        def impl(**kwargs):
            deps = dict()
            for mdep in self.missings:
                deps[mdep] = solve_injection(mdep, kwargs)
            for k, dep in self.kwargs_mapping.items():
                deps[k] = solve_injection(dep, kwargs)
            # logger.info(f"calling function:{self.target_function.__name__}{inspect.signature(self.target_function)}")
            # logger.info(f"src mapping:{self.kwargs_mapping}")
            # logger.info(f"with deps:{deps}")
            return self.target_function(**deps)

        # you have to add a prefix 'provider'""
        return create_function(func_signature=signature, func_impl=impl)

    def dependencies(self) -> Set[str]:
        # ahhhh this recursively demands for injection.
        # we need to distinguish what and what not to recursively inject
        res = set()
        for mdep in self.missings:
            d = extract_dependency(mdep)
            res |= d
            # logger.info(f"deps of missing:{d}")
        for k, dep in self.kwargs_mapping.items():
            d = extract_dependency(dep)
            res |= d
            # logger.info(f"deps of dependency({k}):{d}")
        return res

    # def __str__(self):
    #    return f"""InjectedFunction(target={self.target_function},kwargs_mapping={self.kwargs_mapping})"""


class InjectedPure(Injected[T]):
    __match_args__ = ("value",)

    def __init__(self, value):
        super().__init__()
        self.value = value

    def dependencies(self) -> Set[str]:
        return set()

    def get_provider(self):
        return create_function(func_signature=self.get_signature(), func_impl=lambda: self.value)

    def __str__(self):
        return f"Pure({self.value})"

    def __repr__(self):
        return str(self)


class InjectedByName(Injected[T]):
    __match_args__ = ("name",)

    def __init__(self, name):
        super().__init__()
        self.name = name

    def dependencies(self) -> Set[str]:
        return {self.name}

    def get_provider(self):
        return create_function(func_signature=self.get_signature(), func_impl=lambda **kwargs: kwargs[self.name])

    def __str__(self):
        return f"InjectedByName({self.name})"

    def __repr__(self):
        return str(self)


class ZippedInjected(Injected[Tuple[A, B]]):
    __match_args__ = ("a", "b")

    def __init__(self, a: Injected[A], b: Injected[B]):
        super().__init__()
        assert isinstance(a, Injected), f"got {type(a)} for a"
        assert isinstance(b, Injected), f"got {type(b)} for b"
        self.a = a
        self.b = b

    def dependencies(self) -> Set[str]:
        return set(self.a.dependencies() | self.b.dependencies())

    def get_provider(self):
        def impl(**kwargs):  # can we pickle this though?
            a_kwargs = {k: kwargs[k] for k in self.a.dependencies()}
            b_kwargs = {k: kwargs[k] for k in self.b.dependencies()}
            a = self.a.get_provider()(**a_kwargs)
            b = self.b.get_provider()(**b_kwargs)
            return a, b

        signature = self.get_signature()
        # logger.info(f"created signature:{signature} for ZippedInjected")
        return create_function(func_signature=signature, func_impl=impl)


class MZippedInjected(Injected):
    __match_args__ = ("srcs",)

    def __init__(self, *srcs: Injected):
        super().__init__()
        self.srcs = [Injected.ensure_injected(item) for item in srcs]
        assert all(isinstance(s, Injected) for s in self.srcs), self.srcs

    def dependencies(self) -> Set[str]:
        res = set()
        for s in self.srcs:
            res |= s.dependencies()

        return res

    def get_provider(self):
        def impl(**kwargs):  # can we pickle this though?
            each_kwargs = [{k: kwargs[k] for k in s.dependencies()} for s in self.srcs]
            res = []
            for s in self.srcs:
                r = s.get_provider()(**{k: kwargs[k] for k in s.dependencies()})
                res.append(r)
            return tuple(res)

        signature = self.get_signature()
        # from loguru import logger
        # logger.info(f"created signature:{signature} for MZippedInjected")
        return create_function(func_signature=signature, func_impl=impl)


def _injected_factory(**targets: Injected):
    def _impl(f):
        return Injected.partial(f, **targets)

    return _impl


class InjectedWithDefaultDesign(Injected):
    __match_args__ = ('src', 'default_design_path')

    def __init__(self, src: Injected, default_design_path: str):
        super().__init__()
        self.src = src  # why does this receive 4 play buttons?
        self.default_design_path: str = default_design_path

    def dependencies(self) -> Set[str]:
        return self.src.dependencies()

    def get_provider(self):
        return self.src.get_provider()


def with_default(design_path: str):
    def _impl(f):
        return InjectedWithDefaultDesign(Injected.ensure_injected(f), design_path)

    return _impl


@dataclass
class RunnableInjected(Injected):
    src: Injected
    design_path: str
    working_dir: str

    def dependencies(self) -> Set[str]:
        return self.src.dependencies()

    def get_provider(self):
        return self.src.get_provider()


@dataclass
class PartialInjectedFunction(Injected):
    src: Injected[Callable]

    def __post_init__(self):
        assert isinstance(self.src, Injected), f"src:{self.src} is not an Injected"

    def __call__(self, *args, **kwargs) -> DelegatedVar:
        return self.src.proxy(*args, **kwargs)

    def dependencies(self) -> Set[str]:
        return self.src.dependencies()

    def get_provider(self):
        # logger.warning(f"PartialInjectedFunction.get_provider() is called here.")
        return self.src.get_provider()

    def __hash__(self):
        return hash(self.src)


def add_viz_metadata(metadata: Dict[str, Any]):
    def impl(tgt: Injected):
        if not hasattr(tgt, '__viz_metadata__'):
            tgt.__viz_metadata__ = dict()
        tgt.__viz_metadata__.update(metadata)
        return tgt

    return impl



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Optional, List, Any, Callable

from returns.maybe import Maybe, Nothing, Some

from pinjected import Injected, injected_function
from returns import maybe as raybe

from pinjected.di.graph import IObjectGraph, Providable


def maybe__or__(self, other):
    match self:
        case Some(x):
            return self
        case raybe.Nothing:
            return other


def maybe_filter(self: Maybe, flag_to_keep):
    match self.map(flag_to_keep):
        case Some(True):
            return self
        case _:
            return Nothing


Maybe.__or__ = maybe__or__
Maybe.filter = maybe_filter


class ResourceManifest(ABC):
    """
    determines who can access this resource.
    determines any metadata of this resource.
    """
    pass


class RequesterManifest(ABC):
    """
    determines who is requesting a resource
    metadata of a requester.
    """
    pass


@dataclass(frozen=True)
class ResourcePathManifest(ResourceManifest):
    path: str


@dataclass(frozen=True)
class RequesterPathManifest(RequesterManifest):
    path: str


@dataclass(frozen=True)
class DirectRequesterManifest(RequesterManifest):
    pass


REQUESTER_NO_MANIFEST = DirectRequesterManifest()


@dataclass(frozen=True)
class BindingKey:
    name: str
    manifest: ResourceManifest


@dataclass(frozen=True)
class Binding:
    # a binding uses this manifest to request other dependencies
    src: Injected
    manifest: RequesterManifest


@dataclass(frozen=True)
class Blueprint:
    bindings: Dict[BindingKey, Binding]
    pass


@dataclass
class PermissionManager:
    def is_allowed(self, requester: RequesterManifest, resource: ResourceManifest) -> bool:
        return namespace_permission_rule(requester, resource)
        # return True

    def prioritize(self, manifests: List[ResourceManifest], requester: RequesterManifest) -> List[ResourceManifest]:
        return manifests


class HasPath(ABC):
    __match_args__ = ("path",)

    @classmethod
    def __subclasshook__(cls, C):
        return not hasattr(C, "path")


def namespace_permission_rule(requester: RequesterManifest, resource: ResourceManifest) -> bool:
    """
    check for a full path of the resource and a full path of the requester.
    if a requester lives inside a resource's module, then it is allowed.
    :param requester:
    :param resource:
    :return:
    """
    match requester, resource:
        case (HasPath(path=req_path), HasPath(path=res_path)):
            res_mod = '.'.join(res_path.split(".")[:-1])
            if req_path.startswith(res_mod):
                return True
            else:
                return False
        case _:
            return False


GLOBAL_BINDINGS: Dict[BindingKey, Binding] = dict()


class IBPScope(ABC):
    @abstractmethod
    def provide(self,
                key: BindingKey,
                provider_func: Callable[[], Any],
                trace: list
                ) -> Optional[Providable]:
        pass


@dataclass
class Resolver:
    blueprint: Blueprint
    permission_manager: PermissionManager

    @staticmethod
    def find_binding(
            bindings: Dict[BindingKey, Binding],
            permission_manager: PermissionManager,
            key: str,
            requester: RequesterManifest
    ):
        matched = [k for k in bindings.keys() if
                   k.name == key and permission_manager.is_allowed(requester, k.manifest)]
        if len(matched) == 0:
            return Nothing
        else:
            matched = Resolver.prioritize_binding_keys(permission_manager, matched, requester)
            return Some(bindings[matched[0]])

    @staticmethod
    def prioritize_binding_keys(permission_manager, keys: List[BindingKey], requester):
        """
        prioritize the binding keys according to the requester
        :param keys:
        :param requester:
        :return:
        """
        m2k = {k.manifest: k for k in keys}
        sorted_manifests = permission_manager.prioritize(list(m2k.keys()), requester)
        sorted_keys = [m2k[m] for m in sorted_manifests]
        return sorted_keys

    def find_allowed_binding_key(self, key: str, requester: RequesterManifest) -> Maybe[Binding]:
        """
        find a bindingkey that matches the key and is allowed by the requester
        :param key:
        :param requester:
        :return:
        """
        return self.find_binding(
            self.blueprint.bindings,
            self.permission_manager,
            key,
            requester
        )

    def find_global_default_binding_key(self, key: str, requester: RequesterManifest) -> Maybe[Binding]:
        return self.find_binding(
            GLOBAL_BINDINGS,
            self.permission_manager,
            key,
            requester
        )

    def resolve(self, key: str, requester: RequesterManifest = REQUESTER_NO_MANIFEST):
        assert isinstance(key, str)
        assert isinstance(requester, RequesterManifest)
        binding = self.find_allowed_binding_key(key, requester) | self.find_global_default_binding_key(key, requester)
        match binding:
            case Some(Binding(src, manifest)):
                src: Injected
                deps = src.dependencies()
                resources = {k: self.resolve(k, manifest) for k in deps}
                return src.get_provider()(**resources)
            case raybe.Nothing:
                raise RuntimeError(f"cannot find a binding for key: {key} using {requester}")
            case _:
                raise RuntimeError("unexpected case: %s" % binding)


@dataclass
class BlueprintGraph(IObjectGraph):

    def provide(self, target: Providable, level: int = 2):
        pass

    def child_session(self, overrides=None) -> "IObjectGraph":
        pass

    @property
    def design(self):
        """
        in order to convert a blueprint into a design,
        we need to first resolve the binding key assignments.
        iterate through all the bindings and find appropriate binding so that it can be resolved later using Design.
        I think we should remove this property from an ObjectGraph interface.
        :return:
        """
        raise NotImplementedError("BlueprintGraph does not have a design and this property will be deprecated")


# now let's make a binding key and a bind!

def binding(func):
    src = injected_function(func)
    # get the module path which this function is called.
    import inspect
    frame = inspect.currentframe()
    caller_frame = frame.f_back
    caller_module = inspect.getmodule(caller_frame)
    caller_path = caller_module.__name__
    key = BindingKey(func.__name__, ResourcePathManifest(caller_path))
    bind = Binding(src, RequesterPathManifest(caller_path))
    GLOBAL_BINDINGS[key] = bind
    return bind


"""
Now we have a resolver.
Next thing we need:
1. a Session to manage the instance lifecycle
2. adding a metadata to an injected function.
3. making a blueprint from a design -> abandone
"""

"""
we can actually use this in later,
however, for my current need,
I think it is okey for me to hack
'injected_funcion' to add a namespace metadata
and then make Design implicitly look for it...
hmm,, however, this has some problem after pickling...
since the GLOBAL_BINDINGS won't be pickled over serialization,
the Design has no way of telling the implicit bindings after deserialization.
A solution is to freeze the design with global_bindings on pickling.
Actually, since the Design cannot hold non unique string key, we still need to use a naming of util functions that are unique.
to overcome this, we can use a func name with full module path as a binding key.
and then, use '__' to let the resolver know that the functions should be implicitly resolved.
"""



from pinjected import instances, Design, Injected
from pinjected.di.proxiable import DelegatedVar
from pinjected.module_var_path import ModuleVarPath
from pinjected.run_helpers.run_injected import run_injected


def get_injected(
        var_path: str,
        design_path: str = None,
        _overrides_: str = None,
        **kwargs
):
    """
    load the injected variable from var_path and run it with a design at design_path.
    If design_path is not provided, it will be inferred from var_path.
    design_path is inferred by looking at the module of var_path for a __meta_design__ attribute.
    This command will ask __meta_design__ to provide 'default_design_paths', and uses the first one.
    if __meta_design__ is not found, it will recursively look for a __meta_design__ attribute in the parent module.
    by default, __meta_design__ is accumulated from all parent modules.
    Therefore, if any parent module has a __meta_design__ attribute with a 'default_design_paths' attribute, it will be used.

    :param var_path: the path to the variable to be injected: e.g. "my_module.my_var"
    :param design_path: the path to the design to be used: e.g. "my_module.my_design"
    :param _overrides_: a string that can be converted to an Design in some way. This will gets concatenated to the design.
    :param kwargs: overrides for the design. e.g. "api_key=1234"

    """
    # TODO parse overrides
    instance_overrides = instances(**kwargs)
    overrides = parse_overrides(_overrides_)
    overrides += instance_overrides
    return run_injected("get", var_path, design_path, return_result=True, overrides=overrides)


def parse_overrides(overrides) -> Design:
    match overrides:
        case str() if ':' in overrides:  # this needs to be a complete call to run_injected, at least, we need to take arguments...
            # hmm at this point, we should just run a script ,right?
            design, var = overrides.split(':')
            resolved = run_injected("get", var, design, return_result=True)
            assert isinstance(resolved, Design), f"expected {design} to be a design, but got {resolved}"
            return resolved
        case str() as path:  # a path of a design/injected
            var = ModuleVarPath(path).load()
            if isinstance(var, Design):
                return var
            elif isinstance(var, (Injected, DelegatedVar)):
                resolved = run_injected("get", path, return_result=True)
                assert isinstance(resolved, Design), f"expected {path} to be a design, but got {resolved}"
                return resolved
        case None:
            return instances()


if __name__ == '__main__':
    import fire

    fire.Fire(get_injected)



import asyncio
import inspect
import threading
from abc import ABCMeta, abstractmethod, ABC
from concurrent.futures import Future
from dataclasses import dataclass, field
from functools import lru_cache
from itertools import chain
from pathlib import Path
from pprint import pformat
from typing import Union, Type, Callable, TypeVar, List, Any, Generic, Set, Dict

from returns.maybe import Nothing, Maybe, Some
from returns.result import safe, Result, Failure, Success
from rich.console import Console
from rich.panel import Panel

from pinjected.di.app_injected import EvaledInjected
from pinjected.di.ast import Expr
from pinjected.di.designed import Designed
from pinjected.di.injected import Injected, InjectedByName, InjectedFunction
from pinjected.di.proxiable import DelegatedVar
from pinjected.di.sessioned import Sessioned
from pinjected.exceptions import DependencyResolutionFailure, DependencyResolutionError
from pinjected.graph_inspection import DIGraphHelper

T = TypeVar("T")


class MissingDependencyException(Exception):
    @staticmethod
    def create_message(deps: List[DependencyResolutionFailure]):
        msgs = [item.explanation_str() for item in deps]
        lines = '\n'.join(msgs)
        return f"Missing dependency. failures:\n {lines}."

    @staticmethod
    def create(deps: List[DependencyResolutionFailure]):
        return MissingDependencyException(MissingDependencyException.create_message(deps))


Providable = Union[str, Type[T], Injected[T], Callable, Designed[T], DelegatedVar[Injected], DelegatedVar[Designed]]


class IObjectGraphFactory(ABC):
    def create(self) -> "IObjectGraph":
        pass


@dataclass
class OGFactoryByDesign(IObjectGraphFactory):
    src: "Design"

    def create(self) -> "IObjectGraph":
        return self.src.to_graph()


class IObjectGraph(metaclass=ABCMeta):
    @abstractmethod
    def provide(self, target: Providable, level: int = 2):
        pass

    def __getitem__(self, item: Providable):
        return self.provide(item, level=3)

    @abstractmethod
    def child_session(self, overrides=None) -> "IObjectGraph":
        """
        bindings that are explicit in the overrides are always recreated.
        implicit bindings are created if it is not created in parent. and will be discarded after session.
        else from parent.
        so, implicit bindings that are instantiated at the moment this function is called will be used if not explicitly overriden.
        invalidation is not implemented yet.
        I want the bindings that are dependent on the overriden keys to be reconstructed.
        :param overrides:
        :return:
        """
        raise NotImplementedError()

    def run(self, f):
        argspec = inspect.getfullargspec(f)
        assert "self" not in argspec.args, f"self in {argspec.args}, of {f}"
        # logger.info(self)
        assert argspec.varargs is None
        kwargs = {k: self.provide(k) for k in argspec.args}
        return f(**kwargs)

    def proxied(self, providable: Providable) -> DelegatedVar[Sessioned]:
        from pinjected.di.sessioned import sessioned_ast_context
        designed = self._providable_to_designed(providable)
        item = Sessioned(self, designed)
        ctx = sessioned_ast_context(self)
        from pinjected.di.ast import Object
        return DelegatedVar(Object(item), ctx)

    def sessioned(self, target: Providable) -> DelegatedVar[Union[object, T]]:
        match target:
            case str():
                return self.sessioned(Injected.by_name(target))
            case Injected():
                return self.sessioned(Designed.bind(target))
            case provider if callable(provider):
                return self.sessioned(Injected.bind(provider))
            case Designed():
                val = SessionValue(self, target)
                ctx = sessioned_value_proxy_context(self, val.session)
                return DelegatedVar(val, ctx)
            case DelegatedVar():
                return self.sessioned(target.eval())
            case _:
                raise TypeError(f"Unknown target:{target} queried for DI.")

    def _providable_to_designed(self, target: Providable):
        match target:
            case str():
                return self._providable_to_designed(Injected.by_name(target))
            case Injected():
                return Designed.bind(target)
            case Designed():
                return target
            case DelegatedVar():
                return self._providable_to_designed(target.eval())
            case provider if callable(provider):
                return self._providable_to_designed(Injected.bind(provider))
            case _:
                raise TypeError(f"Unknown target:{target} queried for DI.")

    @property
    @abstractmethod
    def factory(self) -> IObjectGraphFactory:
        pass

    @property
    @abstractmethod
    def design(self):
        pass

    @property
    @abstractmethod
    def resolver(self) -> "DependencyResolver":
        pass


# using async was not a good Idea since all the function needs to be assumed a coroutine.
# we don't assume the provider function is a coroutine.
@dataclass
class ProvideEvent:
    trace: list[str]  # stack of keys currently being provided
    kind: str  # "provide" or "request"
    data: Any = field(default=None)  # provided data when kind == "provide"


class IScope:
    @abstractmethod
    def provide(self, key, provider_func: Callable[[], Any], trace: List) -> Any:
        pass

    @abstractmethod
    def __contains__(self, item):
        raise NotImplementedError()

    @property
    @abstractmethod
    def trace_logger(self):
        pass

    @staticmethod
    def default_trace_logger(event: ProvideEvent):
        from loguru import logger
        match event:
            case ProvideEvent(trace, "request", data=None):
                logger.info(f"{' -> '.join(trace)}")
            case ProvideEvent(trace, "provide", data=[*items]) if len(items) > 100:
                logger.info(f"{' <- '.join(trace)} = LARGE LIST({len(items)} items)")
            case ProvideEvent(trace, "provide", data=[*items]):
                logger.info(f"{' <- '.join(trace)} = {pformat(items)[:100]}")
            case ProvideEvent(trace, "provide", data=data):
                try:
                    buf = str(data)[:100]
                    logger.info(f"{' <- '.join(trace)} = {buf}")
                except Exception as e:
                    logger.error(f"failed to log {trace} with {e}, while providing {' <- '.join(trace)}")


def trace_string(trace: list[str]):
    # res = "\n"
    # for i, item in enumerate(trace):
    #     res += f"{i * '  '}{item} -> \n"
    # return res
    return f" -> ".join(trace)


@dataclass
class RichTraceLogger:
    from rich.console import Console
    console: Console = field(default_factory=lambda: Console(stderr=True))

    def __call__(self, event: ProvideEvent):
        match event:
            case ProvideEvent(trace, "request", data=None):
                self.console.log(f"provide trace:{trace}")
                self.console.log(Panel(trace_string(trace)))
            case ProvideEvent(trace, "provide", data):
                key = trace[-1]
                self.console.log(Panel(trace_string(trace)))
                self.console.log(Panel(pformat(data)))
                # self.console.log(self.value_table(self.values))


@dataclass
class MScope(IScope):
    cache: dict[str, Any] = field(default_factory=dict)
    _trace_logger: Callable[[ProvideEvent], None] = field(default=None)

    def __post_init__(self):
        self._trace_logger = self._trace_logger or self.default_trace_logger
        # self.trace_logger = self.trace_logger or RichTraceLogger()

    @property
    def trace_logger(self):
        return self._trace_logger

    def __getstate__(self):
        raise NotImplementedError("MScope is not serializable")

    def provide(self, key, provider: Callable[[], Any], trace: List) -> Any:
        # with self.lock:
        self.trace_logger(ProvideEvent(trace, "request"))
        if key in self.cache:
            self.trace_logger(ProvideEvent(trace, "provide", data=self.cache[key]))
            return self.cache[key]
        assert isinstance(trace, list) and all(isinstance(item, str) for item in trace)
        assert trace[-1] == key
        res = provider()
        self.cache[key] = res
        res = self.cache[key]
        self.trace_logger(ProvideEvent(trace, "provide", data=res))
        return res

    def __contains__(self, item):
        return item in self.cache


@dataclass
class MChildScope(IScope):
    parent: IScope
    override_targets: set
    cache: dict[str, Any] = field(default_factory=dict)
    _trace_logger: Callable[[ProvideEvent], None] = field(default=None)

    def __post_init__(self):
        self._trace_logger = self._trace_logger or self.default_trace_logger
        # self.trace_logger = self.trace_logger or RichTraceLogger()

    @property
    def trace_logger(self):
        return self._trace_logger

    def __getstate__(self):
        raise NotImplementedError("MChildScope is not serializable")

    def provide(self, key, provider_func: Callable[[], Any], trace: List) -> Any:
        if key not in self.cache:
            if key in self.override_targets or key not in self.parent:
                # logger.info(f"providing from child:{' -> '.join(trace)}")
                self.trace_logger(ProvideEvent(trace, "request"))
                res = provider_func()
                self.cache[key] = res
                res = self.cache[key]
                self.trace_logger(ProvideEvent(trace, "provide", data=res))
                # logger.info(f"provided from child: {' <- '.join(trace)} = {str(res)[:100]}")
            else:
                self.cache[key] = self.parent.provide(key, provider_func, trace)
        return self.cache[key]

    def __contains__(self, item):
        return item in self.cache or item in self.parent


@dataclass
class OverridingScope(IScope):
    """
    This class overrides a given scope with a given set of keys.
    The overriden values will be returned if asked, instead of the original scope.
    """
    src: IScope
    overrides: Dict[str, Any]

    def provide(self, key, provider_func: Callable[[], Any], trace: List) -> Any:
        if key in self.overrides:
            return self.overrides[key]
        else:
            return self.src.provide(key, provider_func, trace)

    def __contains__(self, item):
        return item in self.overrides or item in self.src


class NoMappingError(Exception):
    def __init__(self, key):
        super().__init__(f"No mapping found for DI:{key}")
        self.key = key


@dataclass
class DependencyResolver:
    """
    okey I want to make a variant that uses IMPLICIT_BINDINGS when the target is not in the mapping.
    what we need is actually a str->Injected mapping.
    """
    src: "Design"

    def _to_injected(self, tgt: Providable):
        return providable_to_injected(tgt)

    def __post_init__(self):
        # gather things to build providers graph
        helper = DIGraphHelper(self.src)
        #
        self.mapping: dict[str, Injected] = helper.total_mappings()

        @lru_cache()
        def _memoized_provider(tgt: str):
            return self.mapping[tgt].get_provider()

        self.memoized_provider = _memoized_provider

        predefined = {"__final_target__"}

        @lru_cache()
        def _memoized_deps(tgt: str, include_dynamic=False):
            if tgt in predefined:
                return []
            if tgt not in self.mapping:
                raise NoMappingError(f"target {tgt} is not in the dependency injection mapping.")

            deps = self.mapping[tgt].dependencies()
            if include_dynamic:
                deps = set(deps) | set(self.mapping[tgt].dynamic_dependencies())
            return deps

        self.memoized_deps = _memoized_deps

    def _dfs(self, tgt: str, visited: set[str] = None, include_dynamic=False):
        if visited is None:
            visited = set()
        if tgt in visited:
            return
        visited.add(tgt)
        deps = self.memoized_deps(tgt, include_dynamic=include_dynamic)
        yield tgt
        for dep in deps:
            yield from self._dfs(dep, visited, include_dynamic=include_dynamic)

    def required_dependencies(self, providable: Providable, include_dynamic=False) -> Set[str]:
        tgt: Injected = self._to_injected(providable)
        return set(chain(*[self._dfs(d, include_dynamic=include_dynamic) for d in tgt.dependencies()]))

    def purified_design(self, providable: Providable):
        from pinjected import providers
        deps = self.required_dependencies(providable, include_dynamic=True)
        deps = {k: self.mapping[k] for k in deps}
        return providers(**deps)

    def _dependency_tree(self, tgt: str, trace: list[str] = None) -> Result[Dict[str, Result], Exception]:
        trace = trace or [tgt]
        try:
            res = dict()
            deps = self.memoized_deps(tgt)
            for d in deps:
                assert d not in trace, f"cycle detected: {d} is requested in {' -> '.join(trace)}"
                res[d] = self._dependency_tree(d, trace + [d])
            return Success(res)
        except NoMappingError as ke:
            from loguru import logger
            # msg = f"failed to find dependency for {tgt} in {' -> '.join(trace)}"
            return Failure(DependencyResolutionFailure(tgt, trace, ke))

            # raise RuntimeError(msg) from ke

    def dependency_tree(self, providable: Providable) -> Result[Dict[str, Result], Exception]:
        match providable:
            case str():
                return Success({providable: self._dependency_tree(providable)})
            case _:
                tgt: Injected = self._to_injected(providable)

                return Success({t: self._dependency_tree(t) for t in tgt.dependencies()})

    @staticmethod
    def unresult_tree(tree: Result[Dict[str, Result], Exception]) -> Dict:
        if isinstance(tree, Failure):
            return dict(error=tree)
        else:
            return {k: DependencyResolver.unresult_tree(v) for k, v in tree.unwrap().items()}

    def find_failures(self, tree: Result[Dict[str, Result], Exception]):
        """
        recursively dig into the tree and find all failures
        :param tree:
        :return:
        """

        def dig(t):
            match t:
                case Success(items):
                    return list(chain(*[dig(item) for item in items.values()]))
                case Failure(e):
                    return [e]

        return dig(tree)

    def sorted_dependencies(self, providable: Providable) -> List[Set[str]]:
        match self._to_injected(providable):
            case InjectedByName(name):
                return list(self._dfs(name))
            case tgt:
                from itertools import chain
                visited = set()
                return list(chain(*[self._dfs(d, visited) for d in tgt.dependencies()]))

    def _provide(self, tgt: str, scope: IScope, trace: list[str] = None):
        if trace is None:
            trace = []
        assert isinstance(tgt, str)
        from loguru import logger

        try:
            deps = [self._provide(d, scope, trace + [d]) for d in self.memoized_deps(tgt)]
        except NoMappingError as ke:
            logger.error(f"failed to find dependency for {tgt} in {' -> '.join(trace)}")
            raise DependencyResolutionError(
                f"failed to find dependency for {tgt} in {' -> '.join(trace)}",
                [DependencyResolutionFailure(ke.key, trace, ke)]
            )

        def provider_impl():
            provider = self.memoized_provider(tgt)
            try:
                res = provider(*deps)
            except Exception as e:
                logger.error(f"failed to provide {tgt} with {deps}.\n {' -> '.join(trace)} \n -> {e}")
                raise e
            return res

        res = scope.provide(tgt, provider_impl, trace)
        return res

    def provide(self, providable: Providable, scope: IScope):
        # I need to make this based on Threaded Future rather than asyncio
        # because asyncio does not support creating new loop in a thread
        # which means that we cannot use asyncio.run in a cooruntine
        tgt: Injected = self._to_injected(providable)
        # TODO I want to bind a special key that is only available in this 'provide' scope.
        scope = OverridingScope(scope, overrides=dict(
            __final_target__=tgt,
        ))

        def provide_injected(tgt: Injected, key: str):
            assert isinstance(tgt, Injected), f"tgt must be Injected. got {tgt} of type {type(tgt)}"
            assert isinstance(key, str), f"key must be str. got {key} of type {type(key)}"
            provider = tgt.get_provider()

            # TODO handle the case where this provider raises an exception

            def get_result():
                deps = tgt.dependencies()
                values = [self._provide(d, scope, [key, d]) for d in tgt.dependencies()]
                kwargs = dict(zip(deps, values))
                try:
                    return provider(**kwargs)
                except Exception as e:
                    logger.error(f"failed to provide {key} with {kwargs} \n -> {e}")
                    raise e
                    # raise RuntimeError(f"failed to provide {key} due to an exception!") from e
                # I think we need to give a unique name to this injected so that the value will be cached
                # check if we are in the loop or not

            return scope.provide(key, get_result, trace=[key])

        match tgt:
            case InjectedByName(key):
                res = self._provide(key, scope, trace=[key])
            case EvaledInjected(value, ast) as e:
                of = ast.origin_frame
                assert not isinstance(of, Expr), f"ast.origin_frame must not be Expr. got {of} of type {type(of)}"
                original = ast.origin_frame.filename + ":" + str(ast.origin_frame.lineno)

                key = Path(ast.origin_frame.filename).name + ":" + str(ast.origin_frame.lineno) + "#" + str(id(tgt))
                # key = f"EvaledInjected#{str(id(tgt))}"
                from loguru import logger
                logger.info(f"naming new key: {key} == {original}")
                res = provide_injected(e, key)
            case InjectedFunction(func, kwargs) as IF if IF.origin_frame is not None:
                frame = IF.origin_frame
                original = frame.filename + ":" + str(frame.lineno)
                key = f"InjectedFunction#{str(id(tgt))}"
                from loguru import logger
                logger.info(f"naming new key: {key} == {original}")
                res = provide_injected(IF, key)
            case DelegatedVar(value, cxt) as dv:
                res = self.provide(dv.eval(), scope)
            case Injected():
                from loguru import logger
                logger.info(f"default injected type:{type(tgt)}")
                provider = tgt.get_provider()
                key = provider.__name__ + "#" + str(id(tgt))
                res = provide_injected(tgt, key)
            case _:
                raise TypeError(f"unhandled providable type:{tgt} with type {type(tgt)}")
        # unless we release the block, the coroutine for the remaining task won't get executed.
        # so we must use async provide where 'session' is used
        # one way to prevent this from happening is to use a thread for each provider.
        return res

    def child(self, session_provider, overrides: 'Design' = None):
        if overrides is None:
            from pinjected import Design
            overrides = Design()
        child_design = self.src + overrides.bind_provider(
            session=session_provider
        )
        child_resolver = DependencyResolver(child_design)
        return child_resolver


def run_coroutine_in_new_thread(coroutine):
    # Future to store the result of the coroutine
    future = Future()

    # Function to run the coroutine in a new event loop
    def run_coroutine():
        from loguru import logger
        logger.info(f"running coroutine in new thread")
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            result = loop.run_until_complete(coroutine)
            future.set_result(result)
        except Exception as e:
            future.set_exception(e)
        finally:
            loop.close()

    # Start the new thread
    t = threading.Thread(target=run_coroutine)
    t.start()
    t.join()

    # Return the result of the coroutine
    return future.result()


@safe
def get_caller_info(level: int):
    caller_frame = inspect.currentframe()
    for _ in range(level):
        caller_frame = caller_frame.f_back
    caller_info = inspect.getframeinfo(caller_frame)
    # caller_info can be None

    file_name = caller_info.filename
    line_number = caller_info.lineno

    return file_name, line_number


def providable_to_injected(tgt: Providable) -> Injected:
    match tgt:
        case str():
            return Injected.by_name(tgt)
        case type():
            return Injected.bind(tgt)
        case Injected():
            return tgt
        case Designed():
            raise TypeError(f"cannot use Designed here, since Designed cannot become an Injected.")
        case DelegatedVar(value, cxt):
            return providable_to_injected(tgt.eval())
        case f if callable(f):
            return Injected.bind(f)
        case _:
            raise TypeError(f"target must be either class or a string or Injected. got {tgt}")


@dataclass
class EventDistributor:
    """
    This class is intended to be a hub for a user to subscribe to events.
    I actually hope use rx.Subjects for this purpose; however, their api changes once a year and I cannot ask user to rely on it.
    Therefore, we just prepare callback system for the user.
    The instance of this class can be retrieved via __pinjected_events__ in the dependency injection.
    """
    callbacks: List[Callable[[...], None]] = field(default_factory=list)
    event_history: list[object] = field(default_factory=list)

    def register(self, callback: Callable[[...], None]):
        self.callbacks.append(callback)
        for event in self.event_history:
            callback(event)

    def unregister(self, callback: Callable[[...], None]):
        self.callbacks.remove(callback)

    def emit(self, event: object):
        self.event_history.append(event)
        for callback in self.callbacks:
            callback(event)


@dataclass
class MyObjectGraph(IObjectGraph):
    _resolver: DependencyResolver
    src_design: "Design"
    scope: IScope
    event_distributor: EventDistributor = field(default_factory=EventDistributor)

    def __post_init__(self):
        assert isinstance(self.resolver, DependencyResolver) or self.resolver is None

    @property
    def resolver(self) -> "DependencyResolver":
        return self._resolver

    @property
    def factory(self) -> IObjectGraphFactory:
        return OGFactoryByDesign(self.src_design)

    @staticmethod
    def root(design: "Design", trace_logger=None) -> "MyObjectGraph":
        distributor = EventDistributor()

        def _trace_logger(event):
            distributor.emit(event)
            if trace_logger is not None:
                trace_logger(event)
            else:
                IScope.default_trace_logger(event)

        scope = MScope(_trace_logger=_trace_logger)
        graph = MyObjectGraph(None, design, scope, distributor)
        design = design.bind_instance(
            session=graph,
            __pinjected_events__=graph.event_distributor,
        )
        resolver = DependencyResolver(design)
        graph._resolver = resolver
        return graph

    def provide(self, target: Providable, level: int = 2):
        """
        :param target:
        :param level: 2 when you are direcly calling. set increased number to show the callee
        :return:
        """
        from loguru import logger
        # I need to get the filename and line number of the caller
        if isinstance(target, Designed):
            return self.child_session(target.design)[target.internal_injected]

        target: Injected = providable_to_injected(target)
        fn, ln = get_caller_info(level).value_or(("unknown_function", "unknown_line"))
        dep_tree = self.resolver.dependency_tree(target)
        dep_tree = DependencyResolver.unresult_tree(dep_tree)
        script = self.design.to_vis_graph().to_python_script(
            root=target,
            design_path="__dummy__.design",
        )
        logger.debug(f'Pseudo code of the DI graph for ({str(target)[:100]}):\n{script}')
        # logger.debug(f"Pseudo code of the DI graph:")
        # console = Console(stderr=True)
        # script = Syntax(script, "python", theme="ansi_dark", line_numbers=True)
        # console.log(script)
        # rich.print(script)

        failures = self.resolver.find_failures(dep_tree)
        if failures:
            logger.error(f"DI failures: \n{pformat(failures)}")
            raise DependencyResolutionError(f"DI failures: \n{pformat(failures)}", failures)
        res = self.resolver.provide(target, self.scope)
        # flattened = list(chain(*self.resolver.sorted_dependencies(target)))
        # resolved = {k:repr(self.resolver.provide(k))[:100] for k in flattened}
        # logger.debug(f"DI blueprint resolution result:\n{pformat(resolved)}")
        return res

    def child_session(self, overrides: "Design" = None, trace_logger=None):
        if overrides is None:
            from pinjected import Design
            overrides = Design()
        child_scope = MChildScope(self.scope, set(overrides.keys()),
                                  _trace_logger=trace_logger or self.scope.trace_logger)
        child_graph = MyObjectGraph(None, self.design + overrides, child_scope)
        child_resolver = self.resolver.child(lambda: child_graph, overrides)
        child_graph._resolver = child_resolver
        return child_graph

    @property
    def design(self):
        return self.src_design

    @property
    def design_with_implicits(self):
        from pinjected import providers
        return providers(**DIGraphHelper(self.design).total_mappings()).unbind('session')

    def __repr__(self):
        return f"MyObjectGraph({self.design})"

    def auto_sync(self, rejector=None):
        return AutoSyncGraph(self, rejector)


@dataclass
class AutoSyncGraph:
    src: IObjectGraph
    rejector: Callable[[Any], bool] = field(default=None)
    """
    if rejector returns true for an item, it will not be awaited
    """

    def __post_init__(self):
        if self.rejector is None:
            self.rejector = lambda x: False

    def __getitem__(self, item):
        item = self.src[item]
        if not self.rejector(item) and inspect.isawaitable(item):
            async def waiter():
                return await item

            # I need to distinguish if it's Var or not...

            return asyncio.run(waiter())
        return item


def sessioned_value_proxy_context(parent: IObjectGraph, session: IObjectGraph):
    from pinjected.di.dynamic_proxy import DynamicProxyContextImpl
    return DynamicProxyContextImpl(
        lambda a: a.value,
        lambda x: SessionValue(
            parent,
            Designed.bind(Injected.pure(x)),
            session
        ),
        "SessionValueProxy"
    )


@dataclass
class SessionValue(Generic[T]):
    """a class that holds a lazy value and the session used for producing the value.
    how can I access session of DelegatedVar[SessionValue]?
    current implementatoin uses accessor to get value and then bypassess getattr for this value.
    so I need to tell the context which vars are supposed to be accessed through DelegatedVar
    actually, we can access session through delegatedvar.value.session.
    since value shows the internal value, while eval() returns the final semantic value.
    """
    parent: IObjectGraph
    designed: Designed[T]
    session: IObjectGraph = field(default=None)
    _cache: Maybe[T] = field(default=Nothing)

    def __post_init__(self):
        if self.session is None:
            self.session = self.parent.child_session(self.designed.design)

    @property
    def value(self) -> Any:
        if self._cache is Nothing:
            self._cache = Some(self.session[self.designed.internal_injected])
        return self._cache.unwrap()



from dataclasses import dataclass, field
from typing import Any, Callable, Iterator

from cytoolz import valmap

from pinjected.di.applicative import Applicative
from pinjected.di.ast import Expr, Call, Attr, GetItem, Object, BiOp
from pinjected.di.proxiable import T, DelegatedVar, IProxyContext


@dataclass
class AstProxyContextImpl(IProxyContext[Expr[T]]):
    eval_impl: Callable[[Expr[T]], T]
    iter_impl: Callable[[Expr[T]], Iterator[Expr[T]]] = field(default=None)
    _alias_name: str = field(default="AstProxy")

    def getattr(self, tgt: Expr[T], name: str):
        return self.pure(getattr(tgt, name))

    def call(self, tgt: Expr[T], *args, **kwargs):
        return self.pure(tgt(*args, **kwargs))

    def pure(self, tgt: Expr[T]) -> "DelegatedVar[T]":
        return DelegatedVar(tgt, self)

    def getitem(self, tgt: Expr[T], key) -> Any:
        return self.pure(tgt[key])

    def eval(self, tgt: Expr[T]):
        return self.eval_impl(tgt)

    def alias_name(self):
        return self._alias_name

    def iter(self, tgt: Expr[T]):
        if self.iter_impl is None:
            raise NotImplementedError("iter() not implemented for this context")
        return self.iter_impl(tgt)

    def dir(self, tgt: Expr[T]):
        return dir(tgt)

    def __str__(self):
        return f"{self._alias_name}Context"

    def __hash__(self):
        return hash(self._alias_name)

    def biop_impl(self, op: str, tgt: Expr[T], other):
        match op:
            case '+':
                return self.pure(tgt + other)
            case '-':
                return self.pure(tgt - other)
            case '*':
                return self.pure(tgt * other)
            case '/':
                return self.pure(tgt / other)
            case '%':
                return self.pure(tgt % other)
            case '**':
                return self.pure(tgt ** other)
            case '<<':
                return self.pure(tgt << other)
            case '>>':
                return self.pure(tgt >> other)
            case '&':
                return self.pure(tgt & other)
            case '^':
                return self.pure(tgt ^ other)
            case '|':
                return self.pure(tgt | other)
            case '//':
                return self.pure(tgt // other)
            case '@':
                return self.pure(tgt @ other)
            case '==':
                return self.pure(tgt == other)
            case _:
                raise NotImplementedError(f"biop {op} not implemented")


def ast_proxy(tgt, cxt=AstProxyContextImpl(lambda x: x)):
    return DelegatedVar(Object(tgt), cxt)


def eval_applicative(expr: Expr[T], app: Applicative[T]) -> T:
    def _eval(expr):
        def eval_tuple(expr):
            return tuple(_eval(i) for i in expr)

        def eval_dict(expr):
            return valmap(_eval, expr)

        def ensure_pure(item):
            match item:
                case item if app.is_instance(item):
                    return item
                case DelegatedVar(Expr() as wrapped, AstProxyContextImpl()):
                    return _eval(wrapped)
                case item:
                    return app.pure(item)

        match expr:
            case Object([*items] as x) if isinstance(x, list):
                t = app.zip(*[ensure_pure(item) for item in items])
                return app.map(t, list)
            case Object(([*items] as x)) if isinstance(x, tuple):
                t = app.zip(*[ensure_pure(item) for item in items])
                return app.map(t, tuple)
            case Object({**items} as x) if isinstance(x, dict):
                values = app.zip(*[ensure_pure(item) for item in items.values()])
                return app.map(values, lambda t: {k: v for k, v in zip(items.keys(), t)})

            case Object(x):
                return ensure_pure(x)
            case Call(Expr() as f, args, kwargs):
                injected_func: "T[Callable]" = _eval(f)
                args = app.zip(*eval_tuple(args))
                kwargs: "T[dict]" = app.dict(**eval_dict(kwargs))

                # now we are all in the world of injected. how can I combine them all?
                # so all the arguments are converted into Injected if not, then combined together
                # so if you are to pass an Injected as an argument, you must wrap it with Injected.pure
                def apply(t):
                    from loguru import logger
                    func, args, kwargs = t
                    return func(*args, **kwargs)

                applied = app.map(app.zip(injected_func, args, kwargs), apply)
                return applied
            case Attr(Expr() as data, str() as attr_name):
                injected_data = _eval(data)
                def try_get_attr(x):
                    try:
                        return getattr(x, attr_name)
                    except AttributeError as e:
                        raise RuntimeError(f"failed to get attribute {attr_name} from {x} in AST:{data}") from e
                return app.map(
                    injected_data,
                    try_get_attr
                )

            case GetItem(Expr() as data, Expr() as key):
                injected_data = _eval(data)
                injected_key = _eval(key)
                return app.map(
                    app.zip(injected_data, injected_key),
                    lambda t: t[0][t[1]]
                )
            # case BiOp('+', Expr() as left, Expr() as right):
            #     injected_left = _eval(left)
            #     injected_right = _eval(right)
            #     return app.map(
            #         app.zip(injected_left, injected_right),
            #         lambda t: t[0] + t[1]
            #     )
            case BiOp(op, Expr() as left, Expr() as right):
                injected_left = _eval(left)
                injected_right = _eval(right)

                def eval_biop(t):
                    x, y = t
                    return eval("x " + op + " y", None, dict(x=x, y=y))

                return app.map(
                    app.zip(injected_left, injected_right),
                    eval_biop
                )
            case _:
                raise RuntimeError(f"unsupported ast found!:{type(expr)},{expr}")

    try:
        return _eval(expr)
    except AttributeError as e:
        raise RuntimeError(f"failed to evaluate {expr}") from e



import abc
import asyncio
import inspect
from abc import ABC
from dataclasses import dataclass, field
from typing import Generic, TypeVar, Dict, Any, Optional, Callable, Awaitable, Union

from loguru import logger

from pinjected import Injected, Design, injected, instance
from pinjected import instances as d_instances
from pinjected import providers as d_providers
from pinjected.di.implicit_globals import IMPLICIT_BINDINGS

T = TypeVar('T')
U = TypeVar('U')


@dataclass
class ProvideContext:
    key: Optional['IBindKey']
    parent: Optional['ProvideContext']

    @property
    def trace(self):
        if self.parent is None:
            return []
        else:
            return self.parent.trace + [self]


class IBindKey(ABC):
    """
    hold a providing key and metadata
    """
    pass


@dataclass(frozen=True)
class StrBindKey(IBindKey):
    name: str


class IBind(Generic[T], ABC):
    """
    hold a provider and metadata
    """

    @abc.abstractmethod
    async def provide(self, cxt: ProvideContext, deps: Dict[IBindKey, Any]) -> T:
        pass

    @abc.abstractmethod
    def dependencies(self) -> set[IBindKey]:
        pass

    def amap(self, async_func):
        assert inspect.iscoroutinefunction(async_func), f"async_func must be a coroutine function, got {async_func}"
        return MappedBind(self, async_func)

    @staticmethod
    def zip(*targets: 'IBind'):
        deps = {d for t in targets for d in t.dependencies()}

        async def impl(cxt, deps: dict):
            tasks = []
            for t in targets:
                tasks.append(t.provide(cxt, {k: deps[k] for k in t.dependencies()}))
            return await asyncio.gather(*tasks)

        return JustBind(impl, deps)

    def map(self, func):
        assert not inspect.iscoroutinefunction(func), f"func must not be a coroutine function, got {func}"

        async def async_func(data):
            return func(data)

        return self.amap(async_func)

    @staticmethod
    def dict(**targets: 'IBind'):
        async def mapper(data):  # data is a tuple of results
            return {k: v for k, v in zip(targets.keys(), data)}

        return IBind.zip(*targets).amap(mapper)

    @staticmethod
    def list(*targets: 'IBind'):
        async def mapper(data):
            return list(data)

        return IBind.zip(*targets).amap(mapper)


@dataclass
class JustBind(IBind[T]):
    impl: Callable[[ProvideContext, Dict[IBindKey, Any]], Awaitable[T]]
    deps: set[IBindKey]

    async def provide(self, cxt: ProvideContext, deps: Dict[IBindKey, Any]) -> T:
        return await self.impl(cxt, deps)

    def dependencies(self) -> set[IBindKey]:
        return self.deps


@dataclass
class StrBind(IBind[T]):
    """
    most basic binding that uses strings as keys.
    """
    impl: Callable[[...], Awaitable[T]]
    deps: set[str]

    def __post_init__(self):
        # assert inspect.iscoroutinefunction(self.impl), f"impl must be an coroutine function, got {self.impl}"
        self.keys = {StrBindKey(d) for d in self.deps}

    async def provide(self, cxt: ProvideContext, deps: Dict[IBindKey, Any]) -> T:
        dep_dict = {d.name: deps[d] for d in self.keys}
        return await self.impl(**dep_dict)

    def dependencies(self) -> set[IBindKey]:
        return {StrBindKey(d) for d in self.deps}

    @classmethod
    def pure(cls, value: T) -> 'StrBind[T]':
        async def impl():
            return value

        return cls(impl, set())

    @classmethod
    def async_bind(cls, func: Callable[..., Awaitable[T]]) -> 'StrBind[T]':
        # assert inspect.iscoroutinefunction(func), f"func must be a coroutine function, got {func}"
        deps = inspect.signature(func).parameters.keys()
        return cls(func, set(deps))

    @classmethod
    def func_bind(cls, func: Callable[..., T]) -> 'StrBind[T]':
        assert not inspect.iscoroutinefunction(func), f"func must be ordinal function, got {func}"

        async def impl(**kwargs):
            return func(**kwargs)

        deps = inspect.signature(func).parameters.keys()
        return cls(impl, set(deps))

    @classmethod
    def bind(cls, func: Callable[..., T]) -> 'StrBind[T]':
        if inspect.iscoroutinefunction(func):
            return cls.async_bind(func)
        else:
            return cls.func_bind(func)


@dataclass
class MappedBind(IBind[U]):
    src: IBind[T]
    async_f: Callable[[T], Awaitable[U]]

    async def provide(self, cxt: ProvideContext, deps: Dict[IBindKey, Any]) -> T:
        data = self.src.provide(cxt, deps)
        return await self.async_f(data)

    def dependencies(self) -> set[IBindKey]:
        return self.src.dependencies()


class _Injected(Generic[T]):
    """

    """


@dataclass(frozen=True)
class Blueprint:
    bindings: Dict[IBindKey, IBind]

    def __add__(self, other):
        match other:
            case Design():
                return self + design_to_blueprint(other)
            case Blueprint():
                return Blueprint(self.bindings | other.bindings)
            case _:
                raise TypeError(f"Blueprint can only be added with Blueprint or Design, got {other}")


@dataclass
class AsyncResolver:
    blueprint: Blueprint
    objects: Dict[IBindKey, Any] = field(default_factory=dict)

    async def provide(self, key: IBindKey, cxt: ProvideContext):
        logger.info(f"providing {key}")
        if key not in self.objects:
            tasks = []
            bind = self.blueprint.bindings[key]
            dep_keys = list(bind.dependencies())
            for key in dep_keys:
                n_cxt = ProvideContext(key=key, parent=cxt)
                tasks.append(self.provide(key, n_cxt))
            res = await asyncio.gather(*tasks)
            deps = dict(zip(dep_keys, res))
            self.objects[key] = await bind.provide(cxt, deps)
        return self.objects[key]


Providable = Union[str, IBindKey, Callable]


@dataclass
class Resolver:
    resolver: AsyncResolver

    def provide(self, tgt: Providable):
        cxt = ProvideContext(key=StrBindKey(tgt), parent=None)
        return asyncio.run(self.resolver.provide(cxt.key, cxt))


def instances(**kwargs):
    return Blueprint({StrBindKey(k): StrBind.pure(v) for k, v in kwargs.items()})


def providers(**kwargs: Callable[..., Union[T, Awaitable[T]]]):
    return Blueprint({StrBindKey(k): StrBind.bind(v) for k, v in kwargs.items()})


def injected_to_str_bind(injected: Injected):
    match injected:
        case object(__is_awaitable__=True):
            return StrBind.async_bind(injected.get_provider())
        case x:
            return StrBind.bind(injected.get_provider())


def design_to_blueprint(design: Design):
    bindings = IMPLICIT_BINDINGS | design.bindings
    return Blueprint({StrBindKey(k): injected_to_str_bind(v.to_injected()) for k, v in bindings.items()})


# %%
"""
TODO make blueprint from a design
by default ,convert the Injected into async func.
"""


@injected
async def z(x, y, /, ):
    await asyncio.sleep(0.1)
    return x + y


@injected
async def test(z, /):
    item = await z()
    logger.info(f"z:{item}")
    return item


@instance  # @instance async def return a CachedAwaitable object
async def test3(z):
    return await z()


# so, test3 is a CachedAwaitable and must be awaited.
# this means that things from original Design must be awaited.
async def provide_test2(test3):
    return await test3


bp = providers(
    test2=provide_test2  # async providers work fine in blueprint
) + d_instances(
    x='x',
    y='y'
) + d_providers(
    z=z,
    test=test,
    test3=test3,
    test4=test3
)


# we can use async_bind on @instance to make it work.
async def wait(t):
    return await t


Resolver(AsyncResolver(bp)).provide('test4')  # calling test returns a coroutine, which is a correct behavior.
# TODO fix things annotated in @instance to return object rather than coroutine.
# also, the providers needs to omit awaits.
# %%
# In order to fully switch to async design, I need to ...
"""
1. implement visualizer -> 3 hours
2. implement partial injection annotator -> 3 hours
3. implement run_main with meta-design -> 1 day
4. implement plugins  -> 3 hours
-> total 2 days of work. hmm....
"""



import loguru
from pathlib import Path

import sys

from dataclasses import dataclass
from loguru import logger
from returns.maybe import maybe, Some

from pinjected import providers, instances
from pinjected.helper_structure import MetaContext
from pinjected.helpers import find_default_design_paths
from pinjected.module_inspector import get_project_root


@dataclass
class ConfigCreationArgs:
    module_path: str
    default_design_path: str = None
    runner_script_path: str = None
    interpreter_path: str = None
    working_dir: str = None

    def to_design(self):
        logger.debug(f"python paths:{sys.path}")
        meta_context = MetaContext.gather_from_path(Path(self.module_path))

        design = providers(
            project_root=lambda module_path: Path(get_project_root(module_path)),
            runner_script_path=lambda: self.runner_script_path or __file__,
            interpreter_path=lambda: self.interpreter_path or sys.executable,
            default_design_paths=lambda: find_default_design_paths(self.module_path, self.default_design_path),
            default_working_dir=lambda project_root: maybe(lambda: self.working_dir)() | Some(
                str(project_root)),
            default_design_path=lambda default_design_paths: default_design_paths[0]
        ) + instances(
            logger=loguru.logger,
            custom_idea_config_creator=lambda x: [],  # type ConfigCreator
            meta_context=meta_context,
            module_path=self.module_path,
        ) + meta_context.accumulated
        logger.info(f"using meta design:{meta_context.accumulated}")
        logger.info(f"custom_idea_config_creator:{design['custom_idea_config_creator']}")
        return design



from dataclasses import dataclass
from typing import List

from pydantic import validator as field_validator

from pinjected import Injected, Designed
from pinjected.di.app_injected import InjectedEvalContext
from pinjected.di.proxiable import DelegatedVar
from pinjected.module_inspector import ModuleVarSpec, inspect_module_for_type


def get_runnables(module_path) -> List[ModuleVarSpec]:
    def accept(name, tgt):
        match (name, tgt):
            case (n, _) if n.startswith("provide"):
                return True
            case (_, Injected()):
                return True
            case (_, Designed()):
                return True
            case (_, DelegatedVar(value, cxt)) if cxt == InjectedEvalContext:
                return True
            case (_, DelegatedVar(value, cxt)):
                return False
            case (_, item) if hasattr(item, "__runnable_metadata__") and isinstance(item.__runnable_metadata__, dict):
                return True
            case _:
                return False

    runnables = inspect_module_for_type(module_path, accept)
    return runnables


@dataclass
class RunnableValue:
    """
    I think it is easier to make as much configuration as possible on this side.
    """
    src: ModuleVarSpec
    design_path: str

    @field_validator('src')
    def validate_src_type(cls, value):
        match value:
            case ModuleVarSpec(Injected(), _):
                return value
            case ModuleVarSpec(Designed(), _):
                return value
            case _:
                raise ValueError(f"src must be an instance of Injected of ModuleVarSpec, but got {value}")

    class Config:
        arbitrary_types_allowed = True



"""
Here, we aim to implement a dependency injection feature for a dataclass, using
pyhton 3.11's dataclass_transform decorator.

The goal is to allow this kind of code:

@injected
class Data:
    dep1: X = iattr
    dep2: Y = iattr
    x: int = field(default=0)
    ...

@injected
def data_user(new_data: class[Data])->Data:
    data = new_data(x=10)
    data.dep1.do_something()
    data.dep2.do_something()
    ...

"""
from dataclasses import dataclass

from pinjected import injected


@dataclass
class Data:
    """
    docs for data
    """
    x: int = 0


class DataFactory:
    def __call__(self, x: int) -> Data:
        """
        This is a hello world example docstring
        :param x: tell me something special
        :return: Data
        """
        return Data(x)


# new_data:type[Data] = Data
@injected
def data_user(new_data: DataFactory,/,msg) -> Data:
    d = new_data(x=0)
    pass


class IDataUser:
    def __call__(self,msg:str):
        """
        This is a hello world example docstring
        :param msg:
        :return:
        """
        pass



from pinjected.ide_supports.create_configs import run_with_meta_context

if __name__ == '__main__':
    """
    An entrypoint for running a injected with __meta_design__ integrated.
    This is used to create IDE configurations, and other metadata related tasks.
    """
    import fire
    fire.Fire(run_with_meta_context)



import abc
from dataclasses import dataclass
from typing import Callable, Generic, TypeVar

from returns.maybe import Maybe, Nothing

from pinjected import Injected
from pinjected.di.metadata.location_data import CodeLocation

T, U = TypeVar("T"), TypeVar("U")


@dataclass
class BindMetadata:
    """
    This is the metadata of a bind.
    """
    code_location: Maybe[CodeLocation] = None
    """
    The location of the binding location, this will be used by the IDE to jump to the binding location.
    """


class Bind(Generic[T], metaclass=abc.ABCMeta):

    def map(self, f: Callable[[T], U]) -> "Bind[U]":
        return MappedBind(self, f)

    @staticmethod
    def provider(provider):
        return InjectedBind(Injected.bind(provider))

    @staticmethod
    def injected(injected: Injected) -> "Bind":
        return InjectedBind(injected)

    @staticmethod
    def instance(instance: object):
        return InjectedBind(Injected.pure(instance))

    @staticmethod
    def clazz(cls: type):
        return InjectedBind(Injected.bind(cls))

    @abc.abstractmethod
    def to_injected(self) -> Injected:
        raise NotImplementedError("to_injected must be implemented")

    def __call__(self, *args, **kwargs):
        return self.to_injected().get_provider()(*args, **kwargs)

    @property
    def metadata(self) -> Maybe[BindMetadata]:
        return Nothing


@dataclass
class InjectedBind(Bind):
    src: Injected
    metadata: Maybe[BindMetadata] = Nothing

    def to_injected(self) -> Injected:
        return self.src


@dataclass
class MappedBind(Bind[U]):
    src: Bind[T]
    f: Callable[[T], U]

    def to_injected(self) -> Injected:
        return self.src.to_injected().map(self.f)

    @property
    def metadata(self) -> Maybe[BindMetadata]:
        return self.src.metadata



from typing import Callable

from returns.maybe import Maybe, Nothing

from pinjected import Injected
from pinjected.di.bindings import BindMetadata, InjectedBind
from pinjected.di.injected import PartialInjectedFunction, InjectedFunction


def update_if_registered(
        func: Injected[Callable],
        updated: Injected[Callable],
        meta: Maybe[BindMetadata] = Nothing,
        binding_key: str = None
):
    match func:
        case PartialInjectedFunction(InjectedFunction(tgt, mapping)):
            res = PartialInjectedFunction(updated)
            from pinjected.di.implicit_globals import IMPLICIT_BINDINGS
            key = binding_key if binding_key is not None else tgt.__name__
            IMPLICIT_BINDINGS[key] = InjectedBind(
                res,
                metadata=meta
            )
            return res
        case _:
            return updated



def notify(text, sound='Glass') -> str:
    """
    pops up a notification with text
    :param text:
    :return: Notification result
    """
    import os
    org = text
    text = text.replace('"', '\\"')
    text = text.replace("'", "")
    script = f"'display notification \"{text}\" with title \"OpenAI notification\" sound name \"{sound}\"'"
    cmd = f"""osascript -e {script} """
    os.system(cmd)
    return f"Notified user with text: {org}"



import inspect
import json
import sys
from dataclasses import asdict
from pathlib import Path
from typing import Mapping

import loguru
from loguru import logger
from returns.maybe import Some

import pinjected
import pinjected.global_configs
from pinjected import injected_function, instances, providers, injected_instance, Injected, Design, instance
from pinjected.di.injected import PartialInjectedFunction, InjectedFunction
from pinjected.di.metadata.location_data import ModuleVarLocation
from pinjected.graph_inspection import DIGraphHelper
from pinjected.helper_structure import MetaContext
from pinjected.helpers import inspect_and_make_configurations, find_default_design_paths
from pinjected.module_inspector import get_project_root
from pinjected.module_var_path import ModuleVarPath
from pinjected.run_config_utils import injected_to_idea_configs

__meta_design__ = instances(
    default_design_paths=["pinjected.ide_supports.create_configs.my_design"]
)

from pinjected.run_helpers.run_injected import run_injected


def run_with_meta_context(
        var_path: str,
        context_module_file_path: str,
        design_path: str = None,
        **kwargs
):
    """
    This is for running a injected with __meta_design__ integrated.
    :param var_path:
    :param context_module_file_path:
    :param design_path:
    :param kwargs:
    :return:
    """
    if not "__meta_design__" in Path(context_module_file_path).read_text():
        raise ValueError(f"{context_module_file_path} does not contain __meta_design__")
    meta_context = MetaContext.gather_from_path(context_module_file_path)
    instance_overrides = instances(
        module_path=context_module_file_path,
        interpreter_path=sys.executable,
        meta_context=meta_context
    ) + instances(**kwargs)
    return run_injected("get", var_path, design_path, return_result=True,
                        overrides=meta_context.accumulated + instance_overrides,
                        notifier=logger.info
                        )


@injected_function
def load_meta_context(
        module_path
):
    meta_context = MetaContext.gather_from_path(module_path)
    return meta_context


my_design = instances(
    logger=loguru.logger,
    runner_script_path=pinjected.run_config_utils.__file__,
    custom_idea_config_creator=lambda spec: [],  # type ConfigCreator
    # this becomes recursive and overflows if we call meta_session inside a parent design...
    default_design_path=None,
    print_to_stdout=True
) + providers(
    inspect_and_make_configurations=inspect_and_make_configurations,
    injected_to_idea_configs=injected_to_idea_configs,
    default_design_paths=lambda module_path, default_design_path: find_default_design_paths(module_path,
                                                                                            default_design_path),
    project_root=lambda module_path: Path(get_project_root(module_path)),
    default_working_dir=lambda project_root: Some(str(project_root)),
)


@injected_function
def create_idea_configurations(
        inspect_and_make_configurations,
        module_path,
        print_to_stdout,
        /,
        wrap_output_with_tag=True
):
    pinjected.global_configs.pinjected_TRACK_ORIGIN = False
    configs = inspect_and_make_configurations(module_path)
    pinjected.global_configs.pinjected_TRACK_ORIGIN = True
    logger.info(f"configs:{configs}")

    # since stdout is contaminated by many other modules,
    # We need to think of other way to pass information.
    # should I use a tempfile?
    # or, maybe we can use a separator... like <pinjected>...</pinjected>
    # so that the caller must parse the output.

    if print_to_stdout:
        data_str = (json.dumps(asdict(configs)))
        if wrap_output_with_tag:
            data_str = f"<pinjected>{data_str}</pinjected>"
        print(data_str)
    else:
        return configs


@injected_instance
def list_injected_keys(
        default_design_paths: list[str]
):
    helper = DIGraphHelper(ModuleVarPath(default_design_paths[0]).load())
    print(json.dumps(sorted(list(helper.total_mappings().keys()))))


def get_filtered_signature(func):
    # Get the original signature
    original_signature = inspect.signature(func)

    # Filter out positional-only parameters
    filtered_params = {
        name: param for name, param in original_signature.parameters.items() if
        param.kind != inspect.Parameter.POSITIONAL_ONLY
    }

    # Create a new signature with the filtered parameters
    new_signature = original_signature.replace(parameters=filtered_params.values())

    # Convert the new signature to a string
    new_signature_str = str(new_signature)
    return func.__name__, new_signature_str


@injected_instance
def list_completions(
        default_design_paths: list[str]
):
    """
    An API to be called from IDE to return completions, based on __meta_design__.
    :param default_design_paths:
    :return:[
        {name,description,tail}
    ]
    """
    if not default_design_paths:
        print(f"<pinjected>{json.dumps([])}</pinjected>")
        return
    helper = DIGraphHelper(ModuleVarPath(default_design_paths[0]).load())
    total_mappings: Mapping[str, Injected] = helper.total_mappings()
    logger.info(f"total_mappings:{total_mappings}")

    def key_to_completion(key):
        tgt = total_mappings[key]
        match tgt:
            case PartialInjectedFunction(InjectedFunction(object(__original__=func), kw_mapping)):
                name, signature = get_filtered_signature(func)
                return dict(
                    name=name,
                    description="injected function",
                    tail=signature
                )

        return dict(
            name=key,
            description=f"injected {key}",  # the type text
            tail=f""  # a function signature
        )

    # so, I want to extract the return type, and the function signature.
    # for this, we can utilize the llm.
    # the bindings are mostly PartialInjectedFunction and its proxy.

    completions = [key_to_completion(key) for key in helper.total_mappings().keys()]
    data_str=json.dumps(completions)
    data_str = "<pinjected>"+data_str+"</pinjected>"
    print(data_str)


@instance
def design_metadata(
        default_design_paths: list[str]
):
    d: Design = ModuleVarPath(default_design_paths[0]).load()
    # we load design, so we need to be careful with not to running things...
    """
    protocol->
    meta:{
        key:str
        location: str
    }
    location:{
        type: (path | coordinates)
        value: qualified_name | file_path:line_no:col_no
    }
    structure = list[meta]
    """
    helper = DIGraphHelper(d)
    metas = []
    for k, bind in helper.total_bindings().items():
        match bind.metadata.bind(lambda m: m.code_location):
            case Some(ModuleVarPath(qualified_name)):
                metas.append(dict(
                    key=k,
                    location=dict(
                        type="path",
                        value=qualified_name
                    )
                ))
            case Some(ModuleVarLocation(fp, line, col)):
                metas.append(dict(
                    key=k,
                    location=dict(
                        type="coordinates",
                        value=f'{fp}:{line}:{col}'
                    )
                ))
    logger.info(f"metas:{metas}")
    print(json.dumps(metas))

# TODO implement a provider of documentations
# TODO implement a provider for jump to definition, s that I can click on the injected variables to see the definition.
# TODO automatically add the injected variable key to the argument list when the user selects to use it.
# TODO show a list of injectable variables in the side bar, or the structure view
# TODO detect a variable assign ment from 'injected' functions and any calls that involve DelegatedVar or injected functions
# TODO make a PartialInjectedAsyncFunction and a proxy for it.
